///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2012-2013 DreamWorks Animation LLC
//
// All rights reserved. This software is distributed under the
// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
//
// Redistributions of source code must retain the above copyright
// and license notice and the following restrictions and disclaimer.
//
// *     Neither the name of DreamWorks Animation nor the names of
// its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// IN NO EVENT SHALL THE COPYRIGHT HOLDERS' AND CONTRIBUTORS' AGGREGATE
// LIABILITY FOR ALL CLAIMS REGARDLESS OF THEIR BASIS EXCEED US$250.00.
//
///////////////////////////////////////////////////////////////////////////

#ifndef OPENVDB_UTIL_LOGGING_HAS_BEEN_INCLUDED
#define OPENVDB_UTIL_LOGGING_HAS_BEEN_INCLUDED

#ifndef OPENVDB_USE_LOG4CPLUS

/// Macros to log messages of various severity levels
/// The message is of the form 'someVar << "some text" << ...'.
#define OPENVDB_LOG_INFO(message)
#define OPENVDB_LOG_WARN(message)          do { std::cerr << message << std::endl; } while (0);
#define OPENVDB_LOG_ERROR(message)         do { std::cerr << message << std::endl; } while (0);
#define OPENVDB_LOG_FATAL(message)         do { std::cerr << message << std::endl; } while (0);
#define OPENVDB_LOG_DEBUG(message)
#define OPENVDB_LOG_DEBUG_RUNTIME(message)

#else // ifdef OPENVDB_USE_LOG4CPLUS

#include <logging_base/logging.h>

#define OPENVDB_LOG_INFO(message)          LOG_INFO(message)
#define OPENVDB_LOG_WARN(message)          LOG_WARN(message)
#define OPENVDB_LOG_ERROR(message)         LOG_ERROR(message)
#define OPENVDB_LOG_FATAL(message)         LOG_FATAL(message)
#ifdef DEBUG
/// Log debugging messages in debug builds only.
#define OPENVDB_LOG_DEBUG(message)         LOG_DEBUG(message)
#else
#define OPENVDB_LOG_DEBUG(message)
#endif
/// Log debugging messages even in non-debug builds.
/// Don't use this in performance-critical code.
#define OPENVDB_LOG_DEBUG_RUNTIME(message) LOG_DEBUG_RUNTIME(message)

#endif // OPENVDB_USE_LOG4CPLUS

#endif // OPENVDB_UTIL_LOGGING_HAS_BEEN_INCLUDED

// Copyright (c) 2012-2013 DreamWorks Animation LLC
// All rights reserved. This software is distributed under the
// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
